#include <array>
#include <cmath>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string.h>
#include <utility>
#include <vector>

#include <petsc/private/dmimpl.h>
#include <petsc/private/dmpleximpl.h> /*I   "petscdmplex.h"   I*/
#include <petscdmplex.h>
#include <petscerror.h>
#include <petscsys.h>
#include <petscts.h>
#include <petscvec.h>

static char help[] = "using petsc.\n\n";

using point = std::array<double, 3>;
using edge = std::pair<int, int>;
using hexahedra = std::array<PetscInt, 8>;

static PetscErrorCode DMPlexCreateDepthStratum(DM dm, DMLabel label,
                                               PetscInt depth, PetscInt pStart,
                                               PetscInt pEnd) {
  IS stratumIS;

  PetscFunctionBegin;
  if (pStart >= pEnd)
    PetscFunctionReturn(0);
  // if (PetscDefined(USE_DEBUG)) {
  //   PetscInt qStart, qEnd, numLevels, level;
  //   PetscBool overlap = PETSC_FALSE;
  //   DMLabelGetNumValues(label, &numLevels);
  //   for (level = 0; level < numLevels; level++) {
  //     DMLabelGetStratumBounds(label, level, &qStart, &qEnd);
  //     if ((pStart >= qStart && pStart < qEnd) ||
  //         (pEnd > qStart && pEnd <= qEnd)) {
  //       overlap = PETSC_TRUE;
  //       break;
  //     }
  //   }
  //   if (!overlap) {
  //     SETERRQ(
  //         PETSC_COMM_SELF, PETSC_ERR_PLIB,
  //         "New depth %D range [%D,%D) overlaps with depth %D range [%D,%D)");
  //   }
  // }

  ISCreateStride(PETSC_COMM_SELF, pEnd - pStart, pStart, 1, &stratumIS);
  DMLabelSetStratumIS(label, depth, stratumIS);
  ISDestroy(&stratumIS);
  PetscFunctionReturn(0);
}

int main(int argc, char *argv[]) {
  std::vector<hexahedra> mesh;
  std::vector<point> meshNodes;
  std::vector<edge> meshEdge;
  point pt;
  hexahedra hex;
  edge edg;

  std::ifstream ifile;
  std::ofstream ofile;
  std::string line;

  double dx, dy, dz, Lx = 1.0, Ly = 1.0, Lz = 1.0;
  int Nx = 5, Ny = 2, Nz = 2;
  // int NT=2*(Nx-1)*(Ny-1);

  int rank, mpi_size;
  PetscInt dim, numFields, numBC, i;
  PetscInt numComp[3], numDof[12], bcField[1];
  PetscErrorCode ierr;

  MPI_Comm mpi_comm;

  PetscInitialize(&argc, &argv, (char *)0, help);
  mpi_comm = PETSC_COMM_WORLD;
  MPI_Comm_rank(mpi_comm, &rank);
  MPI_Comm_size(mpi_comm, &mpi_size);
  PetscInt n_elements = 0;
  PetscInt Nxyz;
  DM dm, dmDist = NULL;

  if (rank == 0) {
    Nxyz = Nx * Ny * Nz;

    dim = 3;
    // dx = Lx / (Nx - 1);
    // dy = Ly / (Ny - 1);
    // dz = Lz / (Nz - 1);

    // // nodes coordinates in vector "meshNodes" as 'point' data type
    // for (int k = 0; k < Nz; k++) {
    //   for (int j = 0; j < Ny; j++)
    //     for (int i = 0; i < Nx; i++) {
    //       pt[0] = i * dx;
    //       pt[1] = j * dy;
    //       pt[2] = k * dz;

    //       meshNodes.push_back(pt);
    //       std::cout << pt[0] << "," << pt[1] << "," << pt[2] << std::endl;
    //     }
    // }

    // std::string permission = "no";

    // if (!strcmp(permission.c_str(), "yes")) {

    //   ofile.open("coordinates.dat");
    //   for (int i = 0; i < Nxyz; i++)
    //     ofile << meshNodes[i][0] << "," << meshNodes[i][1] << ","
    //           << meshNodes[i][2] << std::endl;
    //   ofile.close();
    // }
    std::vector<std::string> tokens;
    ifile.open("coordinates.dat");
    // char a='#';

    for (std::string line; getline(ifile, line);) {
      tokens.clear();
      if (line[0] == '#') {

        std::cout << line << "\n";

        for (std::string line; getline(ifile, line);) {

          std::stringstream ss(line);
          int i = 0;
          for (std::string line; getline(ss, line, ' ');) {
            hex[i] = stoi(line);
            i++;
          }
          std::cout << line << std::endl;
          mesh.push_back(hex);

          n_elements += 1;
        }
      }
    }
  }

  MPI_Bcast(&dim, 1, MPIU_INT, 0, mpi_comm);
  ierr = DMCreate(mpi_comm, &dm);
  CHKERRQ(ierr);
  ierr = DMSetType(dm, DMPLEX);
  CHKERRQ(ierr);

  ierr = DMSetDimension(dm, dim);
  CHKERRQ(ierr);
  ierr = DMPlexSetChart(dm, 0, Nxyz + n_elements);
  CHKERRQ(ierr);

  if (rank == 0) {
    for (int c = 0; c < n_elements; ++c) {
      ierr = DMPlexSetConeSize(dm, c, (PetscInt)mesh[c].size());
      CHKERRQ(ierr);
    }
  }

  DMSetUp(dm);

  if (rank == 0) {
    for (int c = 0; c < n_elements; ++c) {
      ierr = DMPlexSetCone(dm, c, (PetscInt *)mesh[c].data());
      CHKERRQ(ierr);
    }
  }

  ierr = DMPlexSymmetrize(dm);
  CHKERRQ(ierr);

  ierr = DMPlexStratify(dm);
  CHKERRQ(ierr);

  PetscBool interpolate = PETSC_TRUE;
  if (interpolate) {
    DM idm;
    ierr = DMPlexInterpolate(dm, &idm);
    CHKERRQ(ierr);
    ierr = DMDestroy(&dm);
    CHKERRQ(ierr);
    dm = idm;
  }
  ierr = DMPlexDistribute(dm, 0, NULL, &dmDist);
  CHKERRQ(ierr);

  PetscFinalize();

  return 0;
}
